"""Exceptions for models and their parsing."""


class ParseError(Exception):
    """Raise on Retrosheet data parsing errors."""

    def __init__(
        self, looking_for_value: str, raw_value: str, game_line: str | None = None, file_path: str | None = None
    ):
        self.looking_for_value = looking_for_value
        self.raw_value = raw_value
        self.game_line = game_line
        self.file_path = file_path
        message = f"Unable to parse '{looking_for_value}' from '{raw_value}'"
        if game_line:
            message = f"{message} in line '{game_line}'"
        if file_path:
            message = f"{message} from '{file_path}'"

        super().__init__(message)



"""Retrieve, load, and persist retrosheet.org data."""
from collections.abc import Iterator
from dataclasses import dataclass, field
from io import BytesIO
from pathlib import Path
from zipfile import ZipFile

from requests import Session


@dataclass
class RetrosheetClient:
    """retrosheet.org client to retrieve Retrosheet data files.

    Data files: https://www.retrosheet.org/game.htm

    Args:
        base_url: the URL base for the client

    Attributes:
        base_url: the URL base for the client
        _session: internal requests session
    """

    base_url: str = "https://www.retrosheet.org"
    _session: Session = field(init=False)

    def __post_init__(self):
        """Initialize the requests session."""
        self._session = Session()

    def get_zip_archive_of_years_play_by_play_data(self, year: int) -> ZipFile:
        """Get the zip archive of a year's play-by-play data.

        Args:
            year: the year to retrieve play-by-play data for.
        """
        response = self._session.get(f"{self.base_url}/events/{year}eve.zip")
        response.raise_for_status()
        return ZipFile(BytesIO(response.content))


def retrieve_years_play_by_play_files(
    year: int,
    data_dir: Path,
    retrosheet_client: RetrosheetClient | None = None,
    force_download: bool = False,
) -> list[Path]:
    """Retrieve a year's play-by-play files.

    Args:
        year: the year to retrieve play-by-play files for
        data_dir: the dir to retrieve/store play-by-play files from/to
        retrosheet_client: a Retrosheet client
        force_download: do not use existing data and force a new download of the data
    """
    retrosheet_client = retrosheet_client or RetrosheetClient()
    data_files = list(_yield_years_play_by_play_files(data_dir, year))
    if data_files and not force_download:
        return data_files

    data_zip_archive = retrosheet_client.get_zip_archive_of_years_play_by_play_data(year)
    _extract_zip_archive(data_zip_archive, data_dir)
    return list(_yield_years_play_by_play_files(data_dir, year))


def _extract_zip_archive(zip_archive: ZipFile, target_dir: Path) -> None:
    """Extract a zip archive to a target directory.

    If the target directory does not exist, it and its parents will be created.

    Args:
        zip_archive: the zip file to extract
        target_dir: the path of the directory to extract to
    """
    target_dir.mkdir(parents=True, exist_ok=True)
    zip_archive.extractall(target_dir.as_posix())


def _yield_years_play_by_play_files(data_dir: Path, year: int) -> Iterator[Path]:
    """Yield a year's play-by-play files.

    Args:
        data_dir: the directory to yield the files from
        year: the year to retrieve files for
    """
    yield from data_dir.glob(f"{year}*.EVN")  # National League data files
    yield from data_dir.glob(f"{year}*.EVA")  # American League data files
    yield from data_dir.glob(f"{year}*.EVF")  # Federal League data files
    yield from data_dir.glob(f"{year}*.EVR")  # Negro League data files



"""Encapsulates Retrosheet modifiers as part of play data."""

import re
from dataclasses import dataclass
from enum import Enum, auto

from pyretrosheet.models.base import Base
from pyretrosheet.models.exceptions import ParseError
from pyretrosheet.models.play.ignored import trim_ignored_characters


class ModifierType(Enum):
    """Play modifier type, as defined in the Retrosheet spec.

    Retrosheet Spec:

    Play modifiers and explanations
        Each modifier is preceded by / in a play record.
        As always, % indicates one the four bases and $ indicates a fielder.

        AP    appeal play
        BP    pop up bunt
        BG    ground ball bunt
        BGDP  bunt grounded into double play
        BINT  batter interference
        BL    line drive bunt
        BOOT  batting out of turn
        BP    bunt pop up
        BPDP  bunt popped into double play
        BR    runner hit by batted ball
        C     called third strike
        COUB  courtesy batter
        COUF  courtesy fielder
        COUR  courtesy runner
        DP    unspecified double play
        E$    error on $
        F     fly
        FDP   fly ball double play
        FINT  fan interference
        FL    foul
        FO    force out
        G     ground ball
        GDP   ground ball double play
        GTP   ground ball triple play
        IF    infield fly rule
        INT   interference
        IPHR  inside the park home run
        L     line drive
        LDP   lined into double play
        LTP   lined into triple play
        MREV  manager challenge of call on the field
        NDP   no double play credited for this play
        OBS   obstruction (fielder obstructing a runner)
        P     pop fly
        PASS  a runner passed another runner and was called out
        R$    relay throw from the initial fielder to $ with no out made
        RINT  runner interference
        SF    sacrifice fly
        SH    sacrifice hit (bunt)
        TH    throw
        TH%   throw to base %
        TP    unspecified triple play
        UINT  umpire interference
        UREV  umpire review of call on the field
    """

    APPEAL_PLAY = auto()
    POP_UP_BUNT = auto()
    GROUND_BALL_BUNT = auto()
    BUNT_GROUNDED_INTO_DOUBLE_PLAY = auto()
    BATTER_INTERFERENCE = auto()
    LINE_DRIVE_BUNT = auto()
    BATTING_OUT_OF_TURN = auto()
    BUNT_POP_UP = auto()
    BUNT_POPPED_INTO_DOUBLE_PLAY = auto()
    RUNNER_HIT_BY_BATTED_BALL = auto()
    CALLED_THIRD_STRIKE = auto()
    COURTESY_BATTER = auto()
    COURTESY_FIELDER = auto()
    COURTESY_RUNNER = auto()
    UNSPECIFIED_DOUBLE_PLAY = auto()
    ERROR = auto()
    FLY = auto()
    FLY_BALL_DOUBLE_PLAY = auto()
    FAN_INTERFERENCE = auto()
    FOUL = auto()
    FORCE_OUT = auto()
    GROUND_BALL = auto()
    GROUND_BALL_DOUBLE_PLAY = auto()
    GROUND_BALL_TRIPLE_PLAY = auto()
    INFIELD_FLY_RULE = auto()
    INTERFERENCE = auto()
    INSIDE_THE_PARK_HOME_RUN = auto()
    LINE_DRIVE = auto()
    LINED_INTO_DOUBLE_PLAY = auto()
    LINED_INTO_TRIPLE_PLAY = auto()
    MANAGER_CHALLENGE_OF_CALL_ON_THE_FIELD = auto()
    NO_DOUBLE_PLAY_CREDITED_FOR_THIS_PLAY = auto()
    OBSTRUCTION = auto()
    POP_FLY = auto()
    RUNNER_PASSED = auto()
    RELAY_THROW = auto()
    RUNNER_INTERFERENCE = auto()
    SACRIFICE_FLY = auto()
    SACRIFICE_HIT_BUNT = auto()
    THROW = auto()
    UNSPECIFIED_TRIPLE_PLAY = auto()
    UMPIRE_INTERFERENCE = auto()
    UMPIRE_REVIEW_OF_CALL_ON_THE_FIELD = auto()
    HIT_LOCATION = auto()
    # Unsure what these modifiers encode, but they appear frequently enough to define
    B = auto()
    BF = auto()
    BFDP = auto()
    U = auto()
    S = auto()
    RR = auto()
    # These may be typos and the intention was to encode these as their capital letter
    p = auto()
    l = auto()


@dataclass
class Modifier:
    """Encodes a play modifier.

    Args:
        type: the type of modifier, as defined in Retrosheet spec
        hit_location: a hit location, if it exists
        fielder_positions: fielder positions relevant to the modifier, if they exist
        base: a base relevant to the modifier, if it exists
        raw: the raw play modifier
    """

    type: ModifierType
    hit_location: str | None
    fielder_positions: list[int]
    base: Base | None
    raw: str

    @classmethod
    def from_event_modifier(cls, modifier: str) -> "Modifier":
        """Load a modifier from the modifier part of a play's event.

        Args:
            modifier: a modifier part of a play's event
        """
        trimmed_modifier = trim_ignored_characters(modifier)
        modifier_type = _get_modifier_type(trimmed_modifier)
        return cls(
            type=modifier_type,
            hit_location=_get_hit_location(trimmed_modifier, modifier_type),
            fielder_positions=_get_fielder_positions(trimmed_modifier, modifier_type),
            base=_get_base(trimmed_modifier, modifier_type),
            raw=modifier,
        )


def _get_modifier_type(modifier: str) -> ModifierType:
    """Get the modifier type via the raw the modifier.

    Args:
        modifier: a modifier part of a play's event
    """
    # Handle odd case from play in 2004CHA.EVA: 'play,8,0,blakc001,20,BBX,8/!F'
    if modifier == "!F":
        return ModifierType.FLY

    # Handle odd case from play in 2011TEX.EVA: 'play,8,0,swisn001,12,BFCX,5/P!5F'
    if modifier == "P!5F":
        return ModifierType.POP_FLY

    # (\d+.*)? matches hit location which is an optional amount of digits followed by an optional amount
    # of alphabetic characters
    # @TODO: move these to the enum - a tuple of regecies?
    pattern_to_modifier_type = {
        r"AP(\d+.*)?": ModifierType.APPEAL_PLAY,
        r"BP(\d+.*)?": ModifierType.POP_UP_BUNT,
        r"BG(\d+.*)?": ModifierType.GROUND_BALL_BUNT,
        r"BGDP(\d+.*)?": ModifierType.BUNT_GROUNDED_INTO_DOUBLE_PLAY,
        r"BINT(\d+.*)?": ModifierType.BATTER_INTERFERENCE,
        r"BL(\d+.*)?": ModifierType.LINE_DRIVE_BUNT,
        r"BOOT(\d+.*)?": ModifierType.BATTING_OUT_OF_TURN,
        r"BPDP(\d+.*)?": ModifierType.BUNT_POPPED_INTO_DOUBLE_PLAY,
        r"BR(\d+.*)?": ModifierType.RUNNER_HIT_BY_BATTED_BALL,
        r"C(\d+.*)?": ModifierType.CALLED_THIRD_STRIKE,
        r"COUB(\d+.*)?": ModifierType.COURTESY_BATTER,
        r"COUF(\d+.*)?": ModifierType.COURTESY_FIELDER,
        r"COUR(\d+.*)?": ModifierType.COURTESY_RUNNER,
        r"DP(\d+.*)?": ModifierType.UNSPECIFIED_DOUBLE_PLAY,
        r"E\d(\d+.*)?": ModifierType.ERROR,
        r"F(\d+.*)?": ModifierType.FLY,
        r"FDP(\d+.*)?": ModifierType.FLY_BALL_DOUBLE_PLAY,
        r"FINT(\d+.*)?": ModifierType.FAN_INTERFERENCE,
        r"FL(\d+.*)?": ModifierType.FOUL,
        r"FO(\d+.*)?": ModifierType.FORCE_OUT,
        r"G(\d+.*)?": ModifierType.GROUND_BALL,
        r"GDP(\d+.*)?": ModifierType.GROUND_BALL_DOUBLE_PLAY,
        r"GTP(\d+.*)?": ModifierType.GROUND_BALL_TRIPLE_PLAY,
        r"IF(\d+.*)?": ModifierType.INFIELD_FLY_RULE,
        r"INT(\d+.*)?": ModifierType.INTERFERENCE,
        r"IPHR(\d+.*)?": ModifierType.INSIDE_THE_PARK_HOME_RUN,
        r"L(\d+.*)?": ModifierType.LINE_DRIVE,
        r"LDP(\d+.*)?": ModifierType.LINED_INTO_DOUBLE_PLAY,
        r"LTP(\d+.*)?": ModifierType.LINED_INTO_TRIPLE_PLAY,
        r"MREV(\d+.*)?": ModifierType.MANAGER_CHALLENGE_OF_CALL_ON_THE_FIELD,
        r"NDP(\d+.*)?": ModifierType.NO_DOUBLE_PLAY_CREDITED_FOR_THIS_PLAY,
        r"OBS(\d+.*)?": ModifierType.OBSTRUCTION,
        r"P(\d+.*)?": ModifierType.POP_FLY,
        r"PASS(\d+.*)?": ModifierType.RUNNER_PASSED,
        r"R\d.*": ModifierType.RELAY_THROW,
        r"R": ModifierType.RELAY_THROW,
        r"RINT(\d+.*)?": ModifierType.RUNNER_INTERFERENCE,
        r"SF(\d+.*)?": ModifierType.SACRIFICE_FLY,
        r"SH(\d+.*)?": ModifierType.SACRIFICE_HIT_BUNT,
        r"TH(\d)?(\d+.*)?(H)?": ModifierType.THROW,
        r"TP(\d+.*)?": ModifierType.UNSPECIFIED_TRIPLE_PLAY,
        r"UINT(\d+.*)?": ModifierType.UMPIRE_INTERFERENCE,
        r"UREV(\d+.*)?": ModifierType.UMPIRE_REVIEW_OF_CALL_ON_THE_FIELD,
        r"\d+.*": ModifierType.HIT_LOCATION,
        # Not defined in Retrosheet, but appears frequently enough to define
        r"B": ModifierType.B,
        r"B\d+.*": ModifierType.B,
        r"BF": ModifierType.BF,
        r"BFDP": ModifierType.BFDP,
        r"U": ModifierType.U,
        r"U\d.*": ModifierType.U,
        r"S": ModifierType.S,
        r"RR.*": ModifierType.RR,
        r"p": ModifierType.p,
        r"l": ModifierType.l,
    }
    for pattern, modifier_type in pattern_to_modifier_type.items():
        if re.fullmatch(pattern, modifier):
            return modifier_type

    raise ParseError("modifer_type", modifier)


def _get_hit_location(modifier: str, modifier_type: ModifierType) -> str | None:
    """Get the hit location from the modifier, if it exists.

    Args:
        modifier: a modifier part of a play's event
        modifier_type: the modifier type
    """
    hit_location_re = r"[A-Z]+(\d+.*)"
    if modifier_type in [ModifierType.ERROR, ModifierType.RELAY_THROW]:
        hit_location_re = r"[ER]\d(\d+)"
    elif modifier_type == ModifierType.THROW:
        hit_location_re = r"TH\d(\d+)"
    elif modifier_type == ModifierType.HIT_LOCATION:
        hit_location_re = r"(.*)"

    if match := re.fullmatch(hit_location_re, modifier):
        return match.group(1)

    return None


def _get_fielder_positions(modifier: str, modifier_type: ModifierType) -> list[int]:
    """Get the fielder positions from the modifier, if they exist.

    Args:
        modifier: a modifier part of a play's event
        modifier_type: the type of modifier
    """
    if modifier_type in [ModifierType.ERROR, ModifierType.RELAY_THROW]:
        # handle rare-case of RELAY_THROW modifier without fielding positions specified
        if modifier == "R":
            return []

        try:
            # From Retrosheet Spec: 'U' appearing in a fielding sequence indicates the fielder handling the ball is unknown'
            # We encode unknown as fielder postiion 0.
            fielder_positions = []
            for position in re.fullmatch(r"[ER](.*)", modifier).group(1):  # type: ignore
                if position == "U":
                    fielder_positions.append(0)
                elif position in ["S", "R", "B", "M", "D", "N"]:
                    # idk what these positions are supposed to be ¯\_(ツ)_/¯
                    # 1997CHN.EVN: 'play,7,0,mondr002,10,BX,S8/G6M/R6S.1-2'
                    # 1996LAN.EVN: 'play,8,0,goodc001,22,.CCBB1X,S6/BG/6MS/R3BU4.1-3(E6/TH);B-2'
                    # 1996COL.EVN: 'play,3,0,burkj001,10,BX,S98/R89M'
                    # 1996SDN.EVN: 'play,11,0,dunss001,10,BX,8/R8RD'
                    # 1996TOR.EVA: 'play,7,1,grees001,00,X,36(1)/FO/G3/R36N'
                    continue
                elif position == "(":
                    # ignore metadata in parenthesis, e.g. 'R3(TH)'
                    break
                else:
                    fielder_positions.append(int(position))
        except (AttributeError, ValueError) as e:
            raise ParseError("fielder_position", raw_value=modifier) from e
        else:
            return fielder_positions

    return []


def _get_base(modifier: str, modifier_type: ModifierType) -> Base | None:
    """Get the base from the modifier, if it exists.

    Args:
        modifier: a modifier part of a play's event
        modifier_type: the type of modifier
    """
    if modifier_type == ModifierType.THROW:  # noqa: SIM102
        if match := re.fullmatch(r"TH(\d|H)", modifier):
            return Base(match.group(1))

    return None



"""Encapsulates Retrosheet player data."""
from dataclasses import dataclass

from pyretrosheet.models.team import TeamLocation


@dataclass
class Player:
    """A player as defined in Retrosheet.

    Args:
        id: the player's id
        name: the player's name
        team_location: the player's team's location
        batting_order_position: the player's batting order position
        fielding_position: the player's fielding position
        raw: str
    """

    id: str
    name: str
    team_location: TeamLocation
    batting_order_position: int
    fielding_position: int
    is_sub: bool
    raw: str

    @classmethod
    def from_start_or_sub_line(cls, start_or_sub_line: str, is_sub: bool) -> "Player":
        """Load a player from Retrosheet start or sub line.

        Args:
            start_or_sub_line: start or sub line from Retrosheet play-by-play data
                Examples include: 'start,richg001,"Gene Richards",0,1,7', 'sub,votha001,"Austin Voth",1,0,1'
            is_sub: whether the player is coming from a sub line or not
        """
        # handle rare cases where a comma is within a player's name
        for name, new_name in [
            ("George Watkins,", "George Watkins"),
            ("Clyde,Barfoot", "Clyde Barfoot"),
            ("Burgess,Smoky", "Burgess Smoky"),
            ("Richie Scheinblum,", "Richie Scheinblum"),
        ]:
            if name in start_or_sub_line:
                start_or_sub_line = start_or_sub_line.replace(name, new_name)

        # Strange encoding error in 2008SFN.EVN
        # Correct encoding in 2007SFN.EVN: 'sub,frank001,"Kevin Frandsen",1,2,6S'
        if start_or_sub_line == 'sub,frank001,"Kevin Frank001",1,8,11':
            start_or_sub_line = 'sub,frank001,"Kevin Frandsen",1,8,11'

        _, player_id, name, team_location, batting_order_position, fielding_position = start_or_sub_line.split(",")
        return cls(
            id=player_id,
            name=name.replace('"', ""),
            team_location=TeamLocation(int(team_location)),
            batting_order_position=int(batting_order_position),
            fielding_position=int(fielding_position),
            is_sub=is_sub,
            raw=start_or_sub_line,
        )



"""View model data through various filters."""
from collections import defaultdict

from pyretrosheet.models.game import Game
from pyretrosheet.models.play import Play
from pyretrosheet.models.player import Player
from pyretrosheet.models.team import TeamLocation


def get_plays(game: Game, include_home_team: bool = True, include_visiting_team: bool = True) -> list[Play]:
    """Get plays in a game.

    Args:
        game: the game to get plays for
        include_home_team: include plays for the home team
        include_visiting_team: include plays for the visiting team
    """
    plays = []
    for event in game.chronological_events:
        if isinstance(event, Play):
            play = event
            if play.team_location == TeamLocation.HOME and include_home_team:
                plays.append(play)
            if play.team_location == TeamLocation.VISITING and include_visiting_team:
                plays.append(play)
    return plays


def get_players(game: Game, include_home_team: bool = True, include_visiting_team: bool = True) -> list[Player]:
    """Get players in a game.

    Args:
        game: the game to get players for
        include_home_team: include players from the home team
        include_visiting_team: include players from the visiting team
    """
    players = []
    for event in game.chronological_events:
        if isinstance(event, Player):
            player = event
            if player.team_location == TeamLocation.HOME and include_home_team:
                players.append(player)
            if player.team_location == TeamLocation.VISITING and include_visiting_team:
                players.append(player)
    return players


def get_batter_plays(
    game: Game, include_home_team: bool = True, include_visiting_team: bool = True
) -> dict[str, list[Play]]:
    """Get a map of batter id to their plays in a game.

    Args:
        game: the game to get plays from
        include_home_team: include plays from the home team
        include_visiting_team: include plays from the visiting team
    """
    batter_plays = defaultdict(list)
    for play in get_plays(game, include_home_team=include_home_team, include_visiting_team=include_visiting_team):
        batter_plays[play.batter_id].append(play)
    return dict(batter_plays)


def get_inning_plays(
    game: Game, include_home_team: bool = True, include_visiting_team: bool = True
) -> dict[int, list[Play]]:
    """Get a map of inning number to plays in that inning for a game.

    Args:
        game: the game to get plays per inning from
        include_home_team: include plays per inning from the home team
        include_visiting_team: include plays per inning from the visiting team
    """
    inning_plays = defaultdict(list)
    for play in get_plays(game, include_home_team=include_home_team, include_visiting_team=include_visiting_team):
        inning_plays[play.inning].append(play)
    return dict(inning_plays)


def get_team_players(games: list[Game], team_id: str) -> list[Player]:
    """Get the players for a given team among a list of games.

    Args:
        games: games to get players from
        team_id: the retrosheet team id
    """
    players = []
    seen_player_ids = set()
    for game in games:
        if game.home_team_id == team_id:
            include_home_team = True
            include_visiting_team = False
        elif game.visiting_team_id == team_id:
            include_home_team = False
            include_visiting_team = True
        else:
            raise ValueError(f"Could not find {team_id} in game={game.id.raw}")  # noqa: TRY003

        for player in get_players(
            game, include_home_team=include_home_team, include_visiting_team=include_visiting_team
        ):
            if player.id not in seen_player_ids:
                players.append(player)
                seen_player_ids.add(player.id)

    return players



"""Encapsulates a Retrosheet game id."""
import datetime as dt
from dataclasses import dataclass


@dataclass
class GameID:
    """A game ID as defined in Retrosheet.

    Example: id,ATL198304080
        - ATL home team
        - date of April 8th, 1983
        - first game of the day

    Args:
        home_team_id: the home team's id
        date: the date the game was played
        game_number: the number of the game happening on the date (single game (0), first game (1) or second game (2))
        raw: the raw value of the game id
    """

    home_team_id: str
    date: dt.date
    game_number: int
    raw: str

    @classmethod
    def from_id_line(cls, id_line: str) -> "GameID":
        """Load the GameID from a 'id' line.

        Args:
            id_line: the 'id' line
        """
        id_value = id_line.split(",")[1]
        return cls(
            home_team_id=id_value[:3],
            date=dt.date(year=int(id_value[3:7]), month=int(id_value[7:9]), day=int(id_value[9:11])),
            game_number=int(id_value[-1]),
            raw=id_line,
        )



"""Encapsulates Retrosheet play basic description as part of play data."""
import re
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum, auto

from pyretrosheet.models.base import Base


class BatterEvent(Enum):
    """Represents a batter event as part of a play's basic description."""

    UNASSISTED_FIELDED_OUT = auto()
    ASSISTED_FIELDED_OUT = auto()
    GROUNDED_INTO_DOUBLE_PLAY = auto()
    LINED_INTO_DOUBLE_PLAY = auto()
    GROUNDED_INTO_TRIPLE_PLAY = auto()
    LINED_INTO_TRIPLE_PLAY = auto()
    CATCHER_INTERFERENCE = auto()
    SINGLE = auto()
    DOUBLE = auto()
    TRIPLE = auto()
    GROUND_RULE_DOUBLE = auto()
    ERROR = auto()
    FIELDERS_CHOICE = auto()
    ERROR_ON_FOUL_FLY_BALL = auto()
    HOME_RUN_LEAVING_PARK = auto()
    HOME_RUN_INSIDE_PARK = auto()
    HIT_BY_PITCH = auto()
    STRIKEOUT = auto()
    NO_PLAY = auto()
    INTENTIONAL_WALK = auto()
    WALK = auto()


class RunnerEvent(Enum):
    """Represents a runner event as part of a play's basic description."""

    BALK = auto()
    CAUGHT_STEALING = auto()
    DEFENSIVE_INDIFFERENCE = auto()
    OTHER_ADVANCE = auto()
    PASSED_BALL = auto()
    WILD_PITCH = auto()
    PICKED_OFF = auto()
    PICKED_OFF_CAUGHT_STEALING = auto()
    STOLEN_BASE = auto()


@dataclass
class Description:
    """Encodes a basic play description.

    Args:
        batter_event: event performed by the batter, if any
        runner_event: event performed by a runner, if any
        fielder_assists: map of the fielder positions of players to number of assists, if any
        fielder_put_outs: map of the fielder positions of players to number of put outs, if any
        fielder_errors: map of the fielder positions of players to number of errors committed, if any
        fielder_handlers: map of the fielder positions of players to number of times the fielder
            handled the ball for a play that did not result in an out, if any
        put_out_at_base: if a put out is made at a base not normally covered by the fielder,
            the base runner is given explicitly
        stolen_base: a stolen base, if any
        raw: the raw play description
    """

    batter_event: BatterEvent | None
    runner_event: RunnerEvent | None
    fielder_assists: dict[int, int]
    fielder_put_outs: dict[int, int]
    fielder_handlers: dict[int, int]
    fielder_errors: dict[int, int]
    put_out_at_base: Base | None
    stolen_base: Base | None
    raw: str

    @classmethod
    def from_event_description(cls, description: str) -> "Description":
        """Load a description from the description part of a play's event.

        Args:
            description: the description part of a play's event
        """
        batter_event = _get_batter_event(description)
        runner_event = _get_runner_event(description)
        fielding_out_plays = _get_fielding_out_plays(description, batter_event, runner_event)
        fielding_handler_plays = _get_fielding_handler_plays(description, batter_event, runner_event)
        return cls(
            batter_event=batter_event,
            runner_event=runner_event,
            fielder_assists=_get_fielder_assists(fielding_out_plays),
            fielder_put_outs=_get_fielder_put_outs(fielding_out_plays),
            fielder_handlers=_get_fielder_handlers(fielding_handler_plays),
            fielder_errors=_get_fielder_errors(description, batter_event, runner_event),
            put_out_at_base=_get_put_out_at_base(description, batter_event),
            stolen_base=_get_stolen_base(description, runner_event),
            raw=description,
        )


def _get_batter_event(description: str) -> BatterEvent | None:
    """Get the batter event from the description.

    The batter event is encoded at the start of the description, if there is one.

    Args:
        description: the description part of a play's event
    """
    pattern_to_batter_event = {
        r"\d": BatterEvent.UNASSISTED_FIELDED_OUT,
        r"\d{2,}(\(.\))?": BatterEvent.ASSISTED_FIELDED_OUT,
        r"\d+\([123H]\)\d": BatterEvent.GROUNDED_INTO_DOUBLE_PLAY,
        r"\d+\([123H]\)\d\([123H]\)\d": BatterEvent.GROUNDED_INTO_TRIPLE_PLAY,
        r"\d+\(B\)\d+\(.\)": BatterEvent.LINED_INTO_DOUBLE_PLAY,
        r"\d+\(B\)\d+\(.\)\d+\(.\)": BatterEvent.LINED_INTO_TRIPLE_PLAY,
        r"H(R)?": BatterEvent.HOME_RUN_LEAVING_PARK,
        r"H(R)?\d": BatterEvent.HOME_RUN_INSIDE_PARK,
        r"S\d+": BatterEvent.SINGLE,
        r"D\d+": BatterEvent.DOUBLE,
        r"T\d+": BatterEvent.TRIPLE,
        r"E\d": BatterEvent.ERROR,
        r"FLE\d": BatterEvent.ERROR_ON_FOUL_FLY_BALL,
        r"FC\d": BatterEvent.FIELDERS_CHOICE,
        r"C": BatterEvent.CATCHER_INTERFERENCE,
        r"HP": BatterEvent.HIT_BY_PITCH,
        r"DGR": BatterEvent.GROUND_RULE_DOUBLE,
        r"K": BatterEvent.STRIKEOUT,
        r"W": BatterEvent.WALK,
        r"I(W)?": BatterEvent.INTENTIONAL_WALK,
        r"NP": BatterEvent.NO_PLAY,
    }
    for pattern, batter_event in pattern_to_batter_event.items():
        if re.fullmatch(pattern, description):
            return batter_event

    return None


def _get_runner_event(description: str) -> RunnerEvent | None:
    """Get the runner event from the description.

    The runner event is encoded at the start of the description or after a '+' following a
    batter event, if there is one.

    Args:
        description: the description part of a play's event
    """
    # Certain batting events are followed by a '+' and a runner event
    # We remove the batting event and '+' in these cases to ease parsing
    if match := re.fullmatch(r"(K|W|IW)\+(.*)", description):
        description = match.group(2)

    pattern_to_runner_event = {
        r"BK": RunnerEvent.BALK,
        r"CS[23H]\(.*\)": RunnerEvent.CAUGHT_STEALING,
        r"DI": RunnerEvent.DEFENSIVE_INDIFFERENCE,
        r"OA": RunnerEvent.OTHER_ADVANCE,
        r"PB": RunnerEvent.PASSED_BALL,
        r"WP": RunnerEvent.WILD_PITCH,
        r"PO[123H]\(.*\)": RunnerEvent.PICKED_OFF,
        r"POCS[123H]\(.*\)": RunnerEvent.PICKED_OFF_CAUGHT_STEALING,
        r"SB[23H]": RunnerEvent.STOLEN_BASE,
    }
    for pattern, runner_event in pattern_to_runner_event.items():
        if re.fullmatch(pattern, description):
            return runner_event

    return None


def _get_fielding_out_plays(
    description: str, batter_event: BatterEvent | None, runner_event: RunnerEvent | None
) -> list[str]:
    """Get the fielding plays resulting in outs.

    Plays in this context is a string that contains the fielding positions of the fielders involved in the out.

    Args:
        description: the description part of a play's event
        batter_event: the batting event, if it exists
        runner_event: the runner event, if it exists
    """
    fielding_out_plays: list[str] = []
    match batter_event:
        case BatterEvent.UNASSISTED_FIELDED_OUT:
            fielding_out_plays.append(re.fullmatch(r"(\d)", description).group(1))  # type: ignore

        case BatterEvent.ASSISTED_FIELDED_OUT:
            fielding_out_plays.append(re.fullmatch(r"(\d+).*", description).group(1))  # type: ignore

        case BatterEvent.GROUNDED_INTO_DOUBLE_PLAY:
            match = re.fullmatch(r"(\d+)\(.\)(\d+)", description)
            fielding_out_plays.extend(match.group(g) for g in [1, 2])  # type: ignore

        case BatterEvent.GROUNDED_INTO_TRIPLE_PLAY:
            match = re.fullmatch(r"(\d+)\(.\)(\d+)\(.\)(\d+)", description)
            fielding_out_plays.extend(match.group(g) for g in [1, 2, 3])  # type: ignore

        case BatterEvent.LINED_INTO_DOUBLE_PLAY:
            match = re.fullmatch(r"(\d+)\(.\)(\d+)\(.\)", description)
            fielding_out_plays.extend(match.group(g) for g in [1, 2])  # type: ignore

        case BatterEvent.LINED_INTO_TRIPLE_PLAY:
            match = re.fullmatch(r"(\d+)\(.\)(\d+)\(.\)(\d+)\(.\)", description)
            fielding_out_plays.extend(match.group(g) for g in [1, 2, 3])  # type: ignore

    match runner_event:
        case RunnerEvent.CAUGHT_STEALING | RunnerEvent.PICKED_OFF | RunnerEvent.PICKED_OFF_CAUGHT_STEALING:
            # errors, 'E', does not result in an out so we skip these runner events
            if not re.fullmatch(r".*\(.*E.*\)", description):
                fielding_out_plays.append(re.fullmatch(r".*\((.*)\)", description).group(1))  # type: ignore

    corrected_fielding_out_plays = []
    for play in fielding_out_plays:
        # ! represents an exceptional play, which we can ignore here
        corrected_part = play
        if "!" in play:
            corrected_part = play.replace("!", "")
        corrected_fielding_out_plays.append(corrected_part)

    return corrected_fielding_out_plays


def _get_fielding_handler_plays(
    description: str, batter_event: BatterEvent | None, runner_event: RunnerEvent | None
) -> list[str]:
    """Get fielding handler plays (plays that did not result in error or outs).

    Plays in this context is a string that contains the fielding positions of the fielders involved in the handling
    of the play.

    Args:
        description: the description part of a play's event
        batter_event: the batting event, if it exists
        runner_event: the runner event, if it exists
    """
    fielding_handler_plays: list[str] = []
    match batter_event:
        case BatterEvent.SINGLE | BatterEvent.DOUBLE | BatterEvent.TRIPLE | BatterEvent.FIELDERS_CHOICE | BatterEvent.HOME_RUN_INSIDE_PARK:
            match = re.fullmatch(r"(S|D|T|FC|H|HR)(\d+)", description)
            fielding_handler_plays.append(match.group(2))  # type: ignore

    match runner_event:
        case RunnerEvent.CAUGHT_STEALING | RunnerEvent.PICKED_OFF | RunnerEvent.PICKED_OFF_CAUGHT_STEALING:
            match = re.fullmatch(r".*\((.*)\)", description)
            fielder_positions = match.group(1)  # type: ignore
            fielder_positions_not_part_of_an_error = []
            parts = fielder_positions.split("/")
            for part in parts:
                has_error = False
                for i, fielder_position in enumerate(part):
                    if fielder_position == "E" or part[i - 1] == "E":
                        has_error = True
                        continue

                    fielder_positions_not_part_of_an_error.append(fielder_position)

                # no outs would occur if there is an error
                if fielder_positions_not_part_of_an_error and has_error:
                    fielding_handler_plays.append("".join(fielder_positions_not_part_of_an_error))

    return fielding_handler_plays


def _get_fielder_assists(fielding_out_plays: list[str]) -> dict[int, int]:
    """Get a map of fielder positions and the number of assists they made on the play.

    Args:
        fielding_out_plays: plays where fielding outs occurred
    """
    fielder_assists: defaultdict[int, int] = defaultdict(int)
    for play in fielding_out_plays:
        if len(play) == 0:
            continue

        for fielder_position in play[:-1]:
            fielder_assists[int(fielder_position)] += 1

    return dict(fielder_assists)


def _get_fielder_put_outs(fielding_out_plays: list[str]) -> dict[int, int]:
    """Get a map of fielder positions and the number of put outs they made on the play.

    Args:
        fielding_out_plays: plays where fielding outs occurred
    """
    fielder_put_outs: defaultdict[int, int] = defaultdict(int)
    for play in fielding_out_plays:
        put_out_position = int(play[-1])
        fielder_put_outs[put_out_position] += 1

    return dict(fielder_put_outs)


def _get_fielder_handlers(fielding_handler_plays: list[str]) -> dict[int, int]:
    """Get a map of fielder positions and the number of handling actions they made on the play.

    Args:
        fielding_handler_plays: plays where fielding handling occurred
    """
    fielder_handlers: defaultdict[int, int] = defaultdict(int)
    for play in fielding_handler_plays:
        for fielder_position in play:
            fielder_handlers[int(fielder_position)] += 1

    return dict(fielder_handlers)


def _get_fielder_errors(
    description: str, batter_event: BatterEvent | None, runner_event: RunnerEvent | None
) -> dict[int, int]:
    """Get a map of fielder positions and the number of errors they made on the play.

    Args:
        description: the description part of a play's event
        batter_event: the batting event, if it exists
        runner_event: the runner event, if it exists
    """
    fielder_errors: defaultdict[int, int] = defaultdict(int)
    match batter_event:
        case BatterEvent.ERROR:
            match = re.fullmatch(r"E(\d+)", description)
            for fielder_position in match.group(1):  # type: ignore
                fielder_errors[int(fielder_position)] += 1

        case BatterEvent.ERROR_ON_FOUL_FLY_BALL:
            match = re.fullmatch(r"FLE(\d+)", description)
            for fielder_position in match.group(1):  # type: ignore
                fielder_errors[int(fielder_position)] += 1

    match runner_event:
        case RunnerEvent.CAUGHT_STEALING | RunnerEvent.PICKED_OFF | RunnerEvent.PICKED_OFF_CAUGHT_STEALING:
            if match := re.fullmatch(r".*\((.*E.*)\)", description):
                fielder_positions = match.group(1)
                for i, fielder_position in enumerate(fielder_positions):
                    if fielder_position == "E":
                        # the fielder position following the 'E' is the player that made the error
                        fielder_errors[int(fielder_positions[i + 1])] += 1

    return dict(fielder_errors)


def _get_put_out_at_base(description: str, batter_event: BatterEvent | None) -> Base | None:
    """Get the base of a put out if it's a non-conventional base put out at.

    Args:
        description: the description part of a play's event
        batter_event: the batting event, if it exists
    """
    if batter_event == BatterEvent.ASSISTED_FIELDED_OUT:
        match = re.fullmatch(r"\d+\((.)\)", description)
        if match:
            return Base(match.group(1))

    return None


def _get_stolen_base(description: str, runner_event: RunnerEvent | None) -> Base | None:
    if runner_event == RunnerEvent.STOLEN_BASE:
        return Base(re.fullmatch(r".*SB([23H])", description).group(1))  # type: ignore

    return None



"""Load raw Retrosheet data into models."""
from collections.abc import Iterator
from copy import deepcopy
from functools import cache
from pathlib import Path

from pyretrosheet import retrosheet
from pyretrosheet.models.exceptions import ParseError
from pyretrosheet.models.game import Game

PYRETROSHEET_DIR = Path.home() / ".pyretrosheet"
DEFAULT_DATA_DIR = PYRETROSHEET_DIR / "data"


@cache
def load_games(
    year: int, data_dir: Path | str = DEFAULT_DATA_DIR, force_download: bool = False, basic_info_only: bool = False
) -> list[Game]:
    """Load Retrosheet games for a given year.

    Results are cached since data should not differ between executions.

    Args:
        year: the year to load Retrosheet data for
        data_dir: dir where data will be stored (defaults to '~/.pyretrosheet/data')
        force_download: force a fresh download of the data even if it already exists
        basic_info_only: only populate basic info (game id and participating teams)
            useful for quick game discovery due to less overhead in parsing entire game data
    """
    data_dir = data_dir if isinstance(data_dir, Path) else Path(data_dir)
    data_dir.mkdir(parents=True, exist_ok=True)
    return [
        game
        for play_by_play_file in retrosheet.retrieve_years_play_by_play_files(
            year=year,
            data_dir=Path(data_dir) or DEFAULT_DATA_DIR,
            force_download=force_download,
        )
        for game in _get_games_from_play_by_play_file(play_by_play_file, basic_info_only=basic_info_only)
    ]


def _get_games_from_play_by_play_file(file: Path, basic_info_only: bool = False) -> list[Game]:
    """Get games loaded from a play by play file.

    Args:
        file: the file path to the play by play file
        basic_info_only: only populate basic info (game id and participating teams)
    """
    loaded_games = []
    for games_lines in _iter_game_lines(file.read_text().splitlines()):
        try:
            loaded_games.append(Game.from_game_lines(games_lines, basic_info_only=basic_info_only))
        except ParseError as e:
            raise ParseError(e.looking_for_value, e.raw_value, e.game_line, file.as_posix()) from e

    return loaded_games


def _iter_game_lines(lines: list[str]) -> Iterator[list[str]]:
    """Iterate the lines corresponding to each game in a Retrosheet play-by-play file.

    Args:
        lines: lines of a play-by-play file (includes multiple games in a single file)
    """
    current_game_lines: list[str] = []
    for line in lines:
        if line.startswith("id,") and current_game_lines:
            yield deepcopy(current_game_lines)
            current_game_lines.clear()

        current_game_lines.append(line)

    if current_game_lines:
        yield deepcopy(current_game_lines)



"""Encapsulates Retrosheet advances as part of play data."""
import re
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum

from pyretrosheet.models.base import Base
from pyretrosheet.models.exceptions import ParseError


class RunAccreditation(Enum):
    """Accreditations for a run."""

    UNEARNED_RUN = "UR"
    RBI_CREDITED = "RBI"
    RBI_NOT_CREDITED = "NORBI"
    RBI_NOT_CREDITED_2 = "NR"
    TEAM_UNEARNED_RUN = "TUR"


@dataclass
class Advance:
    """Encodes the advance from one base to another.

    Note this also includes outs.

    Args:
        from_base: the base the player is coming from
        to_base: the base the player is advancing to
        additional_info: some advances encode additional info
        fielder_assists: map of the fielder positions of players to number of assists, if any
        fielder_put_out: the fielder position of the player putting the runner out, if any player does a put out
        fielder_errors: map of the fielder positions of players to number of errors committed, if any
        fielder_handlers: map of the fielder positions of players to number of times the fielder
            handled the ball for a play that did not result in a real out, if any
        is_out: the advance results in the runner being out
        is_unearned_run_explicit: if a run is explicitly unearned
        is_rbi_credited_explicit: if a run has an explicit rbi credited
        is_rbi_not_credited_explicit: if a run has an explicit rbi not credited
        is_team_unearned_run_explicit: if a run is explicitly a team unearned run
        raw: the raw advance value
    """

    from_base: Base
    to_base: Base
    additional_info: list[str]
    fielder_assists: list[int]
    fielder_put_out: int | None
    fielder_handlers: list[int]
    fielder_errors: list[int]
    is_out: bool
    is_unearned_run_explicit: bool
    is_rbi_credited_explicit: bool
    is_rbi_not_credited_explicit: bool
    is_team_unearned_run_explicit: bool
    raw: str

    @classmethod
    def from_event_advance(cls, advance: str) -> "Advance":
        """Load an advance from the advance part of a play's event.

        Args:
            advance: the advance part of a play's event
                Examples include: 'B-1', '2-3', '1-2(WP)', '2-H(TUR)', '2-H(E4/TH)(UR)(NR)', '1X2', '1XH(862)'
        """
        from_base, to_base = _get_bases(advance)
        additional_info = _get_additional_info(advance)
        is_out = _is_out(advance)
        return cls(
            from_base=from_base,
            to_base=to_base,
            additional_info=_get_additional_info(advance),
            fielder_assists=_get_fielder_assists(additional_info),
            fielder_put_out=_get_fielder_put_out(additional_info, is_out),
            fielder_handlers=_get_fielder_handlers(additional_info, is_out),
            fielder_errors=_get_fielder_errors(additional_info, is_out),
            is_out=is_out,
            is_unearned_run_explicit=RunAccreditation.UNEARNED_RUN.value in additional_info,
            is_rbi_credited_explicit=RunAccreditation.RBI_CREDITED.value in additional_info,
            is_rbi_not_credited_explicit=(
                RunAccreditation.RBI_NOT_CREDITED.value in additional_info
                or RunAccreditation.RBI_NOT_CREDITED_2.value in additional_info
            ),
            is_team_unearned_run_explicit=RunAccreditation.TEAM_UNEARNED_RUN.value in additional_info,
            raw=advance,
        )


def _get_bases(advance: str) -> tuple[Base, Base]:
    """Get from and to bases from the advance.

    Args:
        advance: the advance description
    """
    match = re.fullmatch(r"([B123H])[-X]([B123H]).*", advance)
    if not match:
        raise ParseError("bases_from_advance", advance)

    return Base(match.group(1)), Base(match.group(2))


def _get_additional_info(advance: str) -> list[str]:
    """Get additional info from an advance.

    Retrosheet description:
        Advances may include additional information in the
        form of one or more parameters specified as a parenthesized strings
        of characters. When more than one parameter is given on an advance
        they are individually parenthesized.

    Args:
        advance: the advance description
    """
    return re.findall(r"\(([^)]+)\)", advance)


def _is_out(advance: str) -> bool:
    """Determine if the runner is out as a result of the advance.

    If there is an error within an out, the out does not occur.

    Retrosheet Spec:
        The error indicator negates the out.

    Args:
        advance: the advance description
    """
    is_out_encoded = bool(re.fullmatch(r"([B123H])X([B123H]).*", advance))
    has_error = bool(re.search(r"\(.*E.*\)", advance))
    return is_out_encoded and not has_error


def _get_fielder_assists(additional_info: list[str]) -> list[int]:
    """Get fielder position numbers of fielders with an assist.

    Note that fielders are still given an assist even if an error follows them and an actual out does not occur,

    Args:
        additional_info: advance additional info
    """
    fielder_assists = []
    for info in _iter_fielding_additional_info(additional_info):
        for i, fielder_position in enumerate(info):
            if fielder_position == "E" or info[i - 1] == "E" or i == len(info) - 1:
                continue

            fielder_assists.append(int(fielder_position))

    return fielder_assists


def _get_fielder_put_out(additional_info: list[str], is_out: bool) -> int | None:
    """Get the fielder position putting the runner out, if they exist.

    Args:
        additional_info: advance additional info
        is_out: if an out occurs
    """
    if not is_out:
        return None

    for info in _iter_fielding_additional_info(additional_info):
        return int(info[-1])

    return None


def _get_fielder_handlers(additional_info: list[str], is_out: bool) -> list[int]:
    """Get fielder position numbers of fielders handling the ball in the case of not an actual out occurring.

    Args:
        additional_info: advance additional info
        is_out: if an out occurs
    """
    if is_out:
        return []

    fielder_handlers = []
    for info in _iter_fielding_additional_info(additional_info):
        for i, fielder_position in enumerate(info):
            if fielder_position == "E" or info[i - 1] == "E":
                continue

            fielder_handlers.append(int(fielder_position))

    return fielder_handlers


def _get_fielder_errors(additional_info: list[str], is_out: bool) -> list[int]:
    """Get fielder position numbers of fielders committing an error on the out.

    Args:
        additional_info: advance additional info
        is_out: if an out occurs
    """
    if is_out:
        return []

    fielder_errors = []
    for info in _iter_fielding_additional_info(additional_info):
        for i, fielder_position in enumerate(info):
            if fielder_position == "E":
                fielder_errors.append(int(info[i + 1]))

    return fielder_errors


def _iter_fielding_additional_info(additional_info: list[str]) -> Iterator[str]:
    """Iterate additional info and any sub-parts (delimited by '/').

    Args:
        additional_info: advance additional info
    """
    # ignore parts that are not useful in fielding calculations
    ignore_parts_re = r"(WP|TH(\d)?|PB|THH|BR|OBS|(B|R)INT|INT|AP)"
    for info in additional_info:
        for part in info.split("/"):
            corrected_part = part
            # ! encodes an exceptional part of a play, of which we can ignore here
            if "!" in part:
                corrected_part = part.replace("!", "")

            if re.fullmatch(ignore_parts_re, corrected_part):
                continue

            # it is unclear what info like `8-2` represents
            if re.fullmatch(r"\d-\d", corrected_part):
                continue

            # it is unclear what info like `5X` represents
            if re.fullmatch(r"\dX", corrected_part):
                continue

            # it is unclear what info like `74H` represents
            if re.fullmatch(r"\d+H", corrected_part):
                continue

            # run accreditations do not encode any fielding info
            try:
                RunAccreditation(corrected_part)
                continue
            except ValueError:
                pass

            yield corrected_part



"""Handle ignored Retrosheet data."""


def trim_ignored_characters(play_or_modifier: str) -> str:
    """Trim ignored characters from play data.

    Args:
        play_or_modifier: a raw Retrosheet play or modifier

    From Retrosheet event file description:
        A play record ending in a number sign, #, indicates that there is some uncertainty in the play.
        Occasionally, a com record may follow providing additional information.
        A play record may also contain exclamation points, "!" indicating an exceptional play and
        question marks "?" indicating some uncertainty in the play,
        "+" indicating a hard hit ball, and "-" for softly hit balls.
        These characters can be safely ignored.
    """
    if play_or_modifier.endswith(("#", "!", "?", "+", "-")):
        return play_or_modifier[:-1]

    return play_or_modifier



"""Represents a Retrosheet base encoding."""
from enum import Enum


class Base(Enum):
    """Retrosheet base encodings."""

    BATTER_AT_HOME = "B"
    FIRST_BASE = "1"
    SECOND_BASE = "2"
    THIRD_BASE = "3"
    HOME = "H"



"""Encapsulates Retrosheet game data."""
from collections.abc import Iterator, Sequence
from dataclasses import dataclass

from pyretrosheet.models.exceptions import ParseError
from pyretrosheet.models.game_id import GameID
from pyretrosheet.models.play import Play
from pyretrosheet.models.player import Player

ChronologicalEvent = Player | Play
ChronologicalEvents = Sequence[ChronologicalEvent]


class GameIDNotFoundError(Exception):
    """Error when unable to find a game's id."""

    def __init__(self, first_game_line: str):
        """Initialize the exception.

        Args:
            first_game_line: the first line of game lines
        """
        super().__init__(f"Unable to find game id for game: {first_game_line=}")


@dataclass
class Game:
    """A game as defined in Retrosheet.

    Args:
        id: the game's id
        info: miscellaneous encoded information like temperature, attendance, umpire names, etc.
        chronological_events: the chronological order of events occurring
        earned_runs: map of player id to earned runs
    """

    id: GameID
    info: dict[str, str]
    chronological_events: ChronologicalEvents
    earned_runs: dict[str, int]

    def __repr__(self) -> str:
        """Pretty representation of a game."""
        lines = [
            "Game(",
            f"  id={self.id},",
            f"  home_team_id={self.home_team_id},",
            f"  visiting_team_id={self.visiting_team_id},",
            f"  num_chronological_events={len(self.chronological_events)},",
            f"  earned_runs={self.earned_runs},",
            ")",
        ]
        return "\n".join(lines)

    @classmethod
    def from_game_lines(cls, game_lines: list[str], basic_info_only: bool = False) -> "Game":
        """Load a game from game lines.

        Args:
            game_lines: game lines from a game
            basic_info_only: only populate basic info (game id and participating teams)
        """
        id_ = None
        info = {}
        chronological_events: ChronologicalEvents = []
        earned_runs = {}
        for i, line in enumerate(game_lines):
            parts = line.split(",")
            try:
                match parts[0]:
                    case "id":
                        id_ = GameID.from_id_line(line)

                    case "info":
                        info[parts[1]] = parts[2]

                    case "start":
                        # start lines mark the end of the id and info lines needed for basic info
                        if basic_info_only:
                            break

                        chronological_events.append(Player.from_start_or_sub_line(line, is_sub=False))

                    case "sub":
                        chronological_events.append(Player.from_start_or_sub_line(line, is_sub=True))

                    case "play":
                        comment_lines = list(_yield_comment_lines_following_play(i, game_lines))
                        chronological_events.append(Play.from_play_line(line, comment_lines))

                    case "data":
                        earned_runs[parts[2]] = int(parts[3])
            except ParseError as e:
                raise ParseError(e.looking_for_value, e.raw_value, line) from e
            except Exception as e:
                raise ParseError("unknown", "unknown", line) from e

        if not id_:
            raise GameIDNotFoundError(game_lines[0])

        return cls(
            id=id_,
            info=info,
            chronological_events=chronological_events,
            earned_runs=earned_runs,
        )

    @property
    def home_team_id(self) -> str:
        """The id of the home team."""
        return self.info["hometeam"]

    @property
    def visiting_team_id(self) -> str:
        """The id of the visiting team."""
        return self.info["visteam"]

    @property
    def pretty_id(self) -> str:
        """Simple, human-readable identifier for the game."""
        return f"{self.id.date.strftime('%Y/%m/%d')} {self.visiting_team_id} @ {self.home_team_id}"


def _yield_comment_lines_following_play(play_line_number: int, game_lines: list[str]) -> Iterator[str]:
    """Yield all comment lines that follow a play line."""
    line_pointer = 1
    while True:
        line = game_lines[play_line_number + line_pointer]
        parts = line.split(",")
        if parts[0] == "com":
            yield line
            line_pointer += 1
        else:
            break



"""Encapsulates Retrosheet play data."""
from dataclasses import dataclass

from pyretrosheet.models.play.description import BatterEvent, RunnerEvent
from pyretrosheet.models.play.event import Event
from pyretrosheet.models.play.modifier import ModifierType
from pyretrosheet.models.team import TeamLocation


@dataclass
class Play:
    """A play as defined in Retrosheet.

    Args:
        inning: the inning the play occurred in
        team_location: the team's location
        batter_id: the player id of the batter
        count: the count on the batter
        pitches: all the pitches to the batter in the plate appearance
        comments: comments regarding the play, from 'com' lines in Retrosheet data
        event: the event describes the play that occurred
        raw: the raw play line
    """

    inning: int
    team_location: TeamLocation
    batter_id: str
    count: str
    pitches: str
    comments: list[str]
    event: Event
    raw: str

    @classmethod
    def from_play_line(cls, play_line: str, comment_lines: list[str] | None) -> "Play":
        """Load a play from a play line.

        Args:
            play_line: line for a play (format: play,inning,home/visitor,player id,count,pitches,event)
                Examples include: 'play,7,0,saboc001,01,CX,8/F78', 'play,1,0,marts002,22,CBCBX,S9/L89S-'
            comment_lines: comment lines that reference the play, if present
        """
        # Handle weird case of 'play,3,1,smitj106,??,,43,2-3' where I think this is an encoding error
        if play_line == "play,3,1,smitj106,??,,43,2-3":
            play_line = "play,3,1,smitj106,??,?,43.2-3"

        _, inning, team_location, batter_id, count, pitches, event = play_line.split(",")
        return cls(
            inning=int(inning),
            team_location=TeamLocation(int(team_location)),
            batter_id=batter_id,
            count=count,
            pitches=pitches,
            comments=[c.split(",")[1] for c in comment_lines or []],
            event=Event.from_play_event(event),
            raw=play_line,
        )

    def is_walk(self) -> bool:
        """Determines if the play resulted in a walk."""
        return self.event.description.batter_event in [
            BatterEvent.WALK,
            BatterEvent.INTENTIONAL_WALK,
        ]

    def is_hit_by_pitch(self) -> bool:
        """Determines if the play resulted in the batter being hit by a pitch."""
        return self.event.description.batter_event == BatterEvent.HIT_BY_PITCH

    def is_sacrifice_fly(self) -> bool:
        """Determines if the play resulted in a sacrifice fly."""
        return any(modifier.type == ModifierType.SACRIFICE_FLY for modifier in self.event.modifiers)

    def is_an_at_bat(self) -> bool:
        """Determines if the play counts as an at bat."""
        is_batter_event_at_bat = self.event.description.batter_event not in [
            BatterEvent.NO_PLAY,
            BatterEvent.CATCHER_INTERFERENCE,
            BatterEvent.ERROR_ON_FOUL_FLY_BALL,
        ]
        is_runner_event_at_bat = self.event.description.runner_event not in [
            RunnerEvent.WILD_PITCH,
            RunnerEvent.CAUGHT_STEALING,
            RunnerEvent.STOLEN_BASE,
            RunnerEvent.OTHER_ADVANCE,
            RunnerEvent.PASSED_BALL,
            RunnerEvent.BALK,
            RunnerEvent.PICKED_OFF,
        ]
        return all(
            [
                is_batter_event_at_bat,
                is_runner_event_at_bat,
                not self.is_walk(),
                not self.is_hit_by_pitch(),
                not self.is_sacrifice_fly(),
            ]
        )

    def is_single(self) -> bool:
        """Determines if the play resulted in a single."""
        return self.event.description.batter_event == BatterEvent.SINGLE

    def is_double(self) -> bool:
        """Determines if the play resulted in a double."""
        return self.event.description.batter_event == BatterEvent.DOUBLE

    def is_triple(self) -> bool:
        """Determines if the play resulted in a triple."""
        return self.event.description.batter_event == BatterEvent.TRIPLE

    def is_home_run(self) -> bool:
        """Determines if the play resulted in a home run."""
        return self.event.description.batter_event in [
            BatterEvent.HOME_RUN_INSIDE_PARK,
            BatterEvent.HOME_RUN_LEAVING_PARK,
        ]

    def is_hit(self) -> bool:
        """Determines if the play resulted in a hit."""
        return any(
            [
                self.is_single(),
                self.is_double(),
                self.is_triple(),
                self.is_home_run(),
            ]
        )

    def batter_gets_on_base(self) -> bool:
        """Determines if the batter on the play gets on base (any base)."""
        return any([self.is_hit(), self.is_walk(), self.is_hit_by_pitch()])



"""Encapsulates Retrosheet event as part of play data."""
import re
from dataclasses import dataclass

from pyretrosheet.models.exceptions import ParseError
from pyretrosheet.models.play.advance import Advance
from pyretrosheet.models.play.description import Description
from pyretrosheet.models.play.ignored import trim_ignored_characters
from pyretrosheet.models.play.modifier import Modifier


@dataclass
class Event:
    """The event of a play as defined in Retrosheet."""

    description: Description
    modifiers: list[Modifier]
    advances: list[Advance]
    raw: str

    @classmethod
    def from_play_event(cls, event: str) -> "Event":
        """Load an event from a play line event value.

        Args:
            event: the event description (last part of a play line)
                Examples include: '8/F78', '9/SF.3-H', 'S9/L9S.2-H;1-3'
        """
        event_trimmed = trim_ignored_characters(event)
        # need to handle this case specifically as to not make the rest of the logic more complex
        # this is the only known case of multiple '.' in a play - likely an encoding error
        if event_trimmed == "FC3/DP/G3S.3XH(32);1X2(8).B-1":
            description_and_modifiers = "FC3/DP/G3S"
            advances = ["3XH(32)", "1X2(8)", "B-1"]
        elif "." in event_trimmed:
            description_and_modifiers, advances_raw = event_trimmed.split(".")
            advances = advances_raw.split(";")
        else:
            description_and_modifiers = event_trimmed
            advances = []

        # there are cases of double slashes which I do not believe adds any extra info - remove them
        description_and_modifiers = description_and_modifiers.replace("//", "/")

        # remove trailing slashes - does not encode anything
        if description_and_modifiers.endswith("/"):
            description_and_modifiers = description_and_modifiers[:-1]

        # I shamelessly used ChatGPT for this pattern since it's difficult to separate the description and modifiers
        # consistently.
        # Examples (description and modifiers split out):
        # A => 'A', []
        # A/B/C => 'A', ['B', 'C']
        # A(1/2) => 'A(1/2)', []
        # A(1/2)/B => 'A(1/2)', ['B']
        parts = re.split(r"/(?![^(]*\))|$", description_and_modifiers)
        try:
            description = parts[0]
            modifiers = [p for p in parts[1:] if p]
        except AttributeError as e:
            raise ParseError("description_and_modifiers", event_trimmed) from e

        return cls(
            description=Description.from_event_description(description),
            modifiers=[Modifier.from_event_modifier(m) for m in modifiers],
            advances=[Advance.from_event_advance(a) for a in advances],
            raw=event,
        )
